import openride as ride

from dataclasses import dataclass
from typing import Optional

import time
import numpy as np


@dataclass
class Hero:
    bounding_box: ride.BoundingBox = ride.BoundingBox(
        ride.Point(0, 0, 0.8), rotation=ride.Rotation(), size=ride.Size(2.3, 0.9, 0.8)
    )
    max_velocity: float = 50.0

    def __post_init__(self):
        self.velocity: float = 0.0
        self.acceleration: float = 0.0
        self.steering: float = 0.0

    def update(self, dt: float):

        steering = self.steering / abs(self.velocity + 1)

        x, y, yaw, v = rear_axle_kinematic_model(
            self.bounding_box.position.x,
            self.bounding_box.position.y,
            self.bounding_box.rotation.yaw,
            self.velocity,
            dt,
            self.acceleration,
            steering,
            self.bounding_box.size.x * 2,
            0.5,
        )

        self.bounding_box.position.x = x
        self.bounding_box.position.y = y
        self.bounding_box.rotation.yaw = yaw
        self.velocity = min((v, self.max_velocity))

    def on_key_press(self, key_event: ride.KeyEvent):

        if key_event.key == "w" or key_event.key == "Up":
            self.acceleration = 25.0
        if key_event.key == "s" or key_event.key == "Down":
            self.acceleration = -25.0

        if key_event.key == "a" or key_event.key == "Left":
            self.steering = 4.0
        if key_event.key == "d" or key_event.key == "Right":
            self.steering = -4.0

    def on_key_release(self, key_event: ride.KeyEvent):
        if key_event.key in ["w", "s", "Up", "Down"]:
            self.acceleration = 0.0
        if key_event.key in ["a", "d", "Left", "Right"]:
            self.steering = 0.0

    def collides(self, colliders: ride.BoundingBoxCollection) -> Optional[int]:
        for index, collider in enumerate(colliders):
            if ride.bird_eye_view_distance(self.bounding_box, collider) == 0.0:
                return index

    def respawn(self):
        self.bounding_box = ride.BoundingBox(
            ride.Point(0, 0, 0.8),
            rotation=ride.Rotation(),
            size=ride.Size(2.3, 0.9, 0.8),
        )
        self.velocity = 0.0
        self.acceleration = 0.0
        self.steering = 0.0


def rear_axle_kinematic_model(x0, y0, yaw0, v0, dt, acceleration, wheel_angle, wheelbase, drag):
    x = x0 + (v0 * np.cos(yaw0) * dt)
    y = y0 + (v0 * np.sin(yaw0) * dt)
    v = v0 + (acceleration * dt)
    v -= v0 * drag * dt
    yaw = yaw0 + (v0 * np.tan(wheel_angle) / wheelbase * dt)
    return x, y, yaw, v


@dataclass
class City:
    size: float = 200
    building_width: float = 25
    road_width: float = 15

    def __post_init__(self):
        self.buildings = ride.BoundingBoxCollection()
        self.traffic_cones = ride.BoundingBoxCollection()
        self.generate_buildings()
        self.generate_traffic_cones()

    def generate_buildings(self, fill_ratio: float = 0.9):
        for x in range(-self.size, self.size, self.building_width + self.road_width):
            for y in range(-self.size, self.size, self.building_width + self.road_width):
                if np.random.random() > fill_ratio:
                    continue
                height = np.random.random() * 15 + 5
                building = ride.BoundingBox(
                    ride.Point(x, y, height),
                    size=ride.Size(self.building_width / 2, self.building_width / 2, height),
                )
                self.buildings.append(building)
        offset = self.size % (self.building_width + self.road_width) + (self.building_width + self.road_width) / 2
        self.buildings = self.buildings.transform(ride.Transform(ride.Point(0, offset)))

    def spawn_traffic_cone(self):
        x = np.random.random() * 2 * self.size - self.size
        y = np.random.random() * 2 * self.size - self.size
        box = ride.BoundingBox(ride.Point(x, y, 0.5), size=ride.Size(0.5, 0.5, 0.5))
        for building in self.buildings:
            if ride.bird_eye_view_distance(box, building) == 0.0:
                self.spawn_traffic_cone()
        self.traffic_cones.append(box)

    def generate_traffic_cones(self, density: float = 0.01):
        amount = int(density * self.size**2)
        for _ in range(amount):
            self.spawn_traffic_cone()


def main():

    hero = Hero()
    city = City(size=220)

    viewer = ride.Viewer(mouse_camera_interactions=False)
    viewer.draw_grid(extent=city.size + 40)
    viewer.hud.toggle_fps_count()
    viewer.interactor.set_callback(ride.Event.KeyPressEvent, hero.on_key_press)
    viewer.interactor.set_callback(ride.Event.KeyReleaseEvent, hero.on_key_release)

    score = 0

    t = time.time()
    while True:
        dt = time.time() - t
        t = time.time()

        if hero.collides(city.buildings):
            hero.respawn()
            viewer.camera.follow(hero.bounding_box, spring=0)
            score -= 5

        collision_index = hero.collides(city.traffic_cones)
        if collision_index:
            city.traffic_cones.pop(collision_index)
            city.spawn_traffic_cone()
            score += 1

        score = max((score, 0))

        hero.update(dt)
        viewer.camera.follow(hero.bounding_box, spring=0.8)

        viewer.draw_model(ride.SubCategory.Car, hero.bounding_box, color=(0.7, 0.6, 0.35))
        [
            viewer.draw_bounding_box(building, color=(0.5, 0.5, 0.6))
            for building in city.buildings
            if viewer.camera.is_in_front(building.position)
        ]
        [
            viewer.draw_model(
                ride.Category.TrafficCone,
                cone,
                color=(0.9, 0.2, 0.1),
                number_polygons=200,
            )
            for cone in city.traffic_cones
            if viewer.camera.is_in_front(cone.position)
        ]

        viewer.hud.add("score", score)
        viewer.update()


if __name__ == "__main__":
    main()



from openride.core.numba.interpolation import interpolate_line_index_at_xy, linear_interpolation
from openride.core.numba.polygon_utils import polygon_contains_point
from openride.core.numba.box_utils import bird_eye_view_box_vertices

from typing import Tuple

import numba
import numpy as np


@numba.njit(cache=True)
def bev_distance_between_points(x1: float, y1: float, x2: float, y2: float) -> float:
    """Minimum distance between two points in bird eye view"""
    return np.sqrt((x1 - x2) ** 2 + (y1 - y2) ** 2)


@numba.njit(cache=True)
def bev_distance_between_box_and_point(
    box_position_x: float,
    box_position_y: float,
    box_size_x: float,
    box_size_y: float,
    box_yaw: float,
    point_x: float,
    point_y: float,
) -> float:
    """Minimum distance between a bounding box and a point in bird eye view"""
    relx = point_x - box_position_x
    rely = point_y - box_position_y
    rotx = relx * np.cos(-box_yaw) - rely * np.sin(-box_yaw)
    roty = relx * np.sin(-box_yaw) + rely * np.cos(-box_yaw)
    dx = max(abs(rotx) - box_size_x, 0)
    dy = max(abs(roty) - box_size_y, 0)
    return np.sqrt(dx * dx + dy * dy)


@numba.njit(cache=True)
def bev_project_point_on_line(vertices: np.ndarray, x: float, y: float) -> Tuple[float, float]:
    """Returns the bird eye view coordinates of the closest point along a polyline from another point"""
    index = interpolate_line_index_at_xy(vertices, x, y)
    vx = linear_interpolation(vertices[:, 0], index)
    vy = linear_interpolation(vertices[:, 1], index)
    return vx, vy


@numba.njit(cache=True)
def bev_distance_between_line_and_point(vertices: np.ndarray, x: float, y: float) -> float:
    """Minimum distance between a polyline and a point in bird eye view"""
    return min(
        [
            point_segment_distance(
                x,
                y,
                vertices[i, 0],
                vertices[i, 1],
                vertices[i + 1, 0],
                vertices[i + 1, 1],
            )
            for i in range(vertices.shape[0] - 1)
        ]
    )


@numba.njit(cache=True)
def bev_distance_between_polygon_and_point(vertices: np.ndarray, x: float, y: float) -> float:
    """Minimum distance between a polygon and a point in bird eye view"""
    if polygon_contains_point(vertices, x, y):
        return 0.0
    return bev_distance_between_line_and_point(np.vstack((vertices, vertices[:0])), x, y)


@numba.njit(cache=True)
def bev_distance_between_line_and_polygon(line_vertices: np.ndarray, poly_vertices: np.ndarray) -> float:
    """Minimum distance between a polyline and a polygon in bird eye view"""
    N1 = line_vertices.shape[0]
    N2 = poly_vertices.shape[0]

    distances = np.zeros((N1 - 1, N2))
    for i1 in range(N1 - 1):
        x11, y11 = line_vertices[i1, :2]
        x12, y12 = line_vertices[(i1 + 1), :2]

        if polygon_contains_point(poly_vertices, x11, y11):
            return 0.0

        for i2 in range(N2):
            x21, y21 = poly_vertices[i2 % N2, :2]
            x22, y22 = poly_vertices[(i2 + 1) % N2, :2]

            distances[i1, i2] = segments_distance(x11, y11, x12, y12, x21, y21, x22, y22)
            if distances[i1, i2] == 0.0:
                return 0.0

    return np.min(distances)


@numba.njit(cache=True)
def bev_distance_between_line_and_box(
    line_vertices: np.ndarray, box_x: float, box_y: float, size_x: float, size_y: float, box_yaw: float
) -> float:
    """Minimum distance between a polyline and a bounding box in bird eye view"""
    return bev_distance_between_line_and_polygon(
        line_vertices, bird_eye_view_box_vertices(box_x, box_y, size_x, size_y, box_yaw)
    )


@numba.njit(cache=True)
def bev_distance_between_polygon_and_box(
    poly_vertices: np.ndarray, box_x: float, box_y: float, size_x: float, size_y: float, box_yaw: float
) -> float:
    """Minimum distance between a polygon and a bounding box in bird eye view"""
    return bev_distance_between_polygons(
        poly_vertices, bird_eye_view_box_vertices(box_x, box_y, size_x, size_y, box_yaw)
    )


@numba.njit(cache=True)
def bev_distance_between_polygons(vertices1: np.ndarray, vertices2: np.ndarray) -> float:
    """Minimum distance between two polygons in bird eye view"""

    for x, y in vertices1[:, :2]:
        if polygon_contains_point(vertices2, x, y):
            return 0.0
    for x, y in vertices2[:, :2]:
        if polygon_contains_point(vertices1, x, y):
            return 0.0

    N1 = vertices1.shape[0]
    N2 = vertices2.shape[0]

    distances = np.zeros((N1, N2))
    for i1 in range(N1):
        x11, y11 = vertices1[i1 % N1, :2]
        x12, y12 = vertices1[(i1 + 1) % N1, :2]

        for i2 in range(N2):
            x21, y21 = vertices2[i2 % N2, :2]
            x22, y22 = vertices2[(i2 + 1) % N2, :2]

            distances[i1, i2] = segments_distance(x11, y11, x12, y12, x21, y21, x22, y22)
            if distances[i1, i2] == 0.0:
                return 0.0

    return np.min(distances)


@numba.njit(cache=True)
def bev_distance_between_boxes(
    box1_x: float,
    box1_y: float,
    size1_x: float,
    size1_y: float,
    box1_yaw: float,
    box2_x: float,
    box2_y: float,
    size2_x: float,
    size2_y: float,
    box2_yaw: float,
) -> float:
    """Minimum distance between two bounding boxes in bird eye view"""
    return bev_distance_between_polygons(
        bird_eye_view_box_vertices(box1_x, box1_y, size1_x, size1_y, box1_yaw),
        bird_eye_view_box_vertices(box2_x, box2_y, size2_x, size2_y, box2_yaw),
    )


@numba.njit(cache=True)
def bev_distance_between_lines(vertices1: np.ndarray, vertices2: np.ndarray) -> float:
    """Minimum distance between two polylines in bird eye view"""
    N1 = vertices1.shape[0]
    N2 = vertices2.shape[0]

    if N1 == 1 and N2 == 1:
        return bev_distance_between_points(vertices1[0, 0], vertices1[0, 1], vertices1[1, 0], vertices1[1, 1])
    elif N1 == 1:
        return bev_distance_between_line_and_point(vertices2, vertices1[0, 0], vertices1[0, 1])
    elif N2 == 1:
        return bev_distance_between_line_and_point(vertices1, vertices2[1, 0], vertices2[1, 1])

    distances = np.zeros((N1 - 1, N2 - 1))
    for i1 in range(N1 - 1):
        x11, y11 = vertices1[i1, :2]
        x12, y12 = vertices1[(i1 + 1), :2]

        for i2 in range(N2 - 1):
            x21, y21 = vertices2[i2, :2]
            x22, y22 = vertices2[(i2 + 1), :2]

            distances[i1, i2] = segments_distance(x11, y11, x12, y12, x21, y21, x22, y22)
            if distances[i1, i2] == 0.0:
                return 0.0

    return np.min(distances)


@numba.njit(cache=True)
def segments_distance(x11: float, y11: float, x12: float, y12: float, x21: float, y21: float, x22: float, y22: float):
    """distance between two segments in the plane:
    one segment is (x11, y11) to (x12, y12)
    the other is   (x21, y21) to (x22, y22)
    """
    if segments_intersect(x11, y11, x12, y12, x21, y21, x22, y22):
        return 0.0
    # try each of the 4 vertices w/the other segment
    distances = (
        point_segment_distance(x11, y11, x21, y21, x22, y22),
        point_segment_distance(x12, y12, x21, y21, x22, y22),
        point_segment_distance(x21, y21, x11, y11, x12, y12),
        point_segment_distance(x22, y22, x11, y11, x12, y12),
    )
    return min(distances)


@numba.njit(cache=True)
def segments_intersect(x11: float, y11: float, x12: float, y12: float, x21: float, y21: float, x22: float, y22: float):
    """whether two segments in the plane intersect:
    one segment is (x11, y11) to (x12, y12)
    the other is   (x21, y21) to (x22, y22)
    """
    dx1 = x12 - x11
    dy1 = y12 - y11
    dx2 = x22 - x21
    dy2 = y22 - y21
    delta = dx2 * dy1 - dy2 * dx1
    if delta == 0:
        return False  # parallel segments
    s = (dx1 * (y21 - y11) + dy1 * (x11 - x21)) / delta
    t = (dx2 * (y11 - y21) + dy2 * (x21 - x11)) / (-delta)
    return (0 <= s <= 1) and (0 <= t <= 1)


@numba.njit(cache=True)
def point_segment_distance(px: float, py: float, x1: float, y1: float, x2: float, y2: float):
    """Minimum distance between two line segments (finite length) in bird eye view"""
    dx = x2 - x1
    dy = y2 - y1
    if dx == dy == 0:  # the segment's just a point
        return np.sqrt((px - x1) ** 2 + (py - y1) ** 2)

    # Calculate the t that minimizes the distance.
    t = ((px - x1) * dx + (py - y1) * dy) / (dx * dx + dy * dy)

    # See if this represents one of the segment's
    # end points or a point in the middle.
    if t < 0:
        dx = px - x1
        dy = py - y1
    elif t > 1:
        dx = px - x2
        dy = py - y2
    else:
        near_x = x1 + t * dx
        near_y = y1 + t * dy
        dx = px - near_x
        dy = py - near_y

    return np.sqrt(dx**2 + dy**2)



import numpy as np
import numba

from typing import Tuple

EPSILON = np.finfo(float).eps


@numba.njit(cache=True)
def spherical_to_cartesian(r: float, theta: float, phi: float) -> Tuple[float, float, float]:
    x = r * np.sin(phi) * np.cos(theta)
    y = r * np.sin(phi) * np.sin(theta)
    z = r * np.cos(phi)
    return x, y, z


@numba.njit(cache=True)
def cartesian_to_spherical(x: float, y: float, z: float) -> Tuple[float, float, float]:
    r = (x**2 + y**2 + z**2) ** 0.5
    theta = np.arctan(y / (x + EPSILON))
    if x < 0:
        theta += np.pi
    phi = np.arccos(z / (r + EPSILON))
    return r, theta, phi



from openride.core.geometry import Geometry
from openride.core.point import Point
from openride.core.numba import transform_vertices

import copy
import numpy as np

from dataclasses import dataclass
from shapely import geometry
from typing import List, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from openride.core.transform import Transform


@dataclass
class Polygon(Geometry):

    vertices: Union[np.ndarray, List[Point]] = np.empty((0, 3))

    def __post_init__(self):

        if isinstance(self.vertices, list):
            points = copy.deepcopy(self.vertices)
            self.vertices = np.zeros((len(points), 3))
            for i, point in enumerate(points):
                self.vertices[i] = point.x, point.y, point.z

        if np.any(self.vertices[0] != self.vertices[-1]):
            self.vertices = np.vstack([self.vertices, self.vertices[0]])

    def to_shapely(self) -> geometry.Polygon:
        return geometry.Polygon(self.vertices)

    def transform(self, transform: "Transform") -> "Polygon":
        return Polygon(transform_vertices(self.vertices, transform.get_matrix()))

    @classmethod
    def from_shapely(cls, polygon: geometry.Polygon) -> "Polygon":
        return cls([Point(*c) for c in polygon.exterior.coords])

    def get_area(self) -> float:
        return self.to_shapely().area



from openride import Point, BoundingBox

import numpy as np


class Camera:
    def __init__(self, viewer):
        self.viewer = viewer
        self._camera = viewer.renderer.GetActiveCamera()
        self._camera.SetClippingRange(0.1, 100000)
        self._camera.SetViewUp(0, 0, 1)

        self.set_position(Point(0, 0, 0))
        self.set_focus(Point(1, 0, 0))

    def set_position(self, position: Point):
        self._camera.SetPosition(position.x, position.y, position.z)

    def set_focus(self, focus: Point):
        self._camera.SetFocalPoint(focus.x, focus.y, focus.z)

    def get_position(self) -> Point:
        return Point(*self._camera.GetPosition())

    def get_focus(self) -> Point:
        return Point(*self._camera.GetFocalPoint())

    def follow(
        self,
        bounding_box: BoundingBox,
        distance: float = 15,
        height: float = 3,
        spring: float = 0.0,
    ):
        self.set_focus(bounding_box.position)

        x = -distance * np.cos(bounding_box.rotation.yaw)
        y = -distance * np.sin(bounding_box.rotation.yaw)
        z = height

        target_position = bounding_box.position + Point(x, y, z)
        self.set_position(spring * self.get_position() + (1 - spring) * target_position)

    def is_in_front(self, point: Point) -> bool:
        position = self.get_position()
        focus = self.get_focus()
        v1 = np.array([point.x - position.x, point.y - position.y, point.z - position.z])
        v2 = np.array([focus.x - position.x, focus.y - position.y, focus.z - position.z])
        return np.dot(v1, v2) >= 0



import time
import vtk


class Hud:
    def __init__(self, viewer, position=(0, 0)):

        self.viewer = viewer
        self.position = position
        self.time = time.time()
        self._elements = {}

        self.text = vtk.vtkTextActor()
        self.text.SetPosition(self.position[0], self.position[1])
        self.text.GetTextProperty().SetFontSize(16)
        self.viewer.renderer.AddActor(self.text)

        self.fps_count: bool = False
        self.cursor_position: bool = False

    def update(self):
        if self.fps_count:
            self.add("fps", int(1 / (time.time() - self.time)))
        if self.cursor_position:
            self.add("cursor", self.viewer.interactor.get_cursor_in_world())
        text = ""
        for element in self._elements:
            text += f"{element}: {str(self._elements[element])}\n"
        self.text.SetInput(text)
        self.time = time.time()

    def toggle_fps_count(self):
        self.fps_count = not self.fps_count

    def toggle_cursor_position(self):
        self.cursor_position = not self.cursor_position

    def add(self, name, value):
        self._elements[name] = value



from openride.core.numba import rotation_matrix, euler_angles

from dataclasses import dataclass

import numpy as np


@dataclass
class Rotation:

    roll: float = 0.0
    pitch: float = 0.0
    yaw: float = 0.0

    def __post_init__(self):
        self.roll = self.roll % (2 * np.pi)
        self.pitch = self.pitch % (2 * np.pi)
        self.yaw = self.yaw % (2 * np.pi)

    def get_matrix(self) -> np.ndarray:
        return rotation_matrix(self.roll, self.pitch, self.yaw)

    @classmethod
    def from_matrix(cls, matrix: np.ndarray) -> "Rotation":
        return cls(*euler_angles(matrix))

    def __add__(self, other: "Rotation") -> "Rotation":
        return Rotation(self.roll + other.roll, self.pitch + other.pitch, self.yaw + other.yaw)

    def __sub__(self, other: "Rotation") -> "Rotation":
        return Rotation(self.roll - other.roll, self.pitch - other.pitch, self.yaw - other.yaw)



from .core.bounding_box import BoundingBox
from .core.categories import Category, SubCategory
from .core.geometry import Geometry
from .core.point import Point
from .core.polygon import Polygon
from .core.polyline import Polyline
from .core.rotation import Rotation
from .core.size import Size
from .core.transform import Transform

from .core.collections.bounding_box_collection import BoundingBoxCollection
from .core.collections.point_cloud import PointCloud

from .core.operations.distance import distance
from .core.operations.bird_eye_view_distance import bird_eye_view_distance
from .core.operations.contains import contains
from .core.operations.bird_eye_view_contains import bird_eye_view_contains

from .viewer.viewer import Viewer
from .viewer.models import get_model_files
from .viewer.interactor import Event, KeyEvent, MouseEvent



from openride.core.collections.collection import Collection
from openride.core.point import Point
from openride.core.transform import Transform
from openride.core.numba import transform_vertices

from typing import List, Union

import numpy as np


class PointCloud(Collection):
    def __init__(self, points: Union[np.ndarray, List[Point]] = np.empty((0, 3))):

        if isinstance(points, list):
            self._raw = np.zeros((len(points), 3))
            for i, point in enumerate(points):
                self._raw[i] = point.x, point.y, point.z

        else:
            self._raw = points

    def __len__(self) -> int:
        return self._raw.shape[0]

    def append(self, obj: Point):
        self._raw = np.vstack((self._raw, np.array((obj.x, obj.y, obj.z))))

    def extend(self, collection: "PointCloud"):
        self._raw = np.vstack((self._raw, collection._raw))

    def pop(self, index: int) -> Point:
        point = self[index]
        self._raw = np.delete(self._raw, index, axis=0)
        return point

    def __getitem__(self, index: int) -> Point:
        return Point(*self._raw[index])

    def get_point_cloud(self) -> np.ndarray:
        return self._raw

    def transform(self, transform: Transform) -> "PointCloud":
        tmp = np.copy(self._raw)
        tmp = transform_vertices(self._raw, transform.get_matrix())
        return PointCloud(tmp)



import vtk

from abc import ABC, abstractclassmethod


class BaseModelLoader(ABC):
    @abstractclassmethod
    def __call__(self, file: str, n_polygons: int = 100000) -> vtk.vtkPolyData:
        """Loads files into vtk.vtkPolyData instances"""


class STLLoader(BaseModelLoader):
    def __init__(self):
        self._loaded_files = {}

    def __call__(self, file: str, n_polygons: int = 100000) -> vtk.vtkPolyData:
        """Loads .stl files into vtk.vtkPolyData instances"""
        if file not in self._loaded_files:
            reader = vtk.vtkSTLReader()
            reader.SetFileName(file)
            reader.Update()
            polydata = reader.GetOutput()

            decimate = vtk.vtkQuadricDecimation()
            decimate.SetInputData(polydata)
            decimate.SetTargetReduction(1.0 - n_polygons / polydata.GetNumberOfPolys())
            decimate.Update()

            self._loaded_files[file] = decimate.GetOutput()

        return self._loaded_files[file]


class OBJLoader(BaseModelLoader):
    def __init__(self):
        self._loaded_files = {}

    def __call__(self, file: str, n_polygons: int = 100000) -> vtk.vtkPolyData:
        """Loads .stl files into vtk.vtkPolyData instances"""
        if file not in self._loaded_files:
            reader = vtk.vtkOBJReader()
            reader.SetFileName(file)
            reader.Update()
            polydata = reader.GetOutput()

            decimate = vtk.vtkQuadricDecimation()
            decimate.SetInputData(polydata)
            decimate.SetTargetReduction(1.0 - n_polygons / polydata.GetNumberOfPolys())
            decimate.Update()

            self._loaded_files[file] = decimate.GetOutput()

        return self._loaded_files[file]


class ModelLoader(BaseModelLoader):
    def __init__(self):
        self.loaders = {
            "stl": STLLoader(),
            "obj": OBJLoader(),
        }

    def __call__(self, file: str, n_polygons: int = 100000) -> vtk.vtkPolyData:
        loader = self.loaders.get(file.split(".")[-1])
        if loader is None:
            raise NotImplementedError
        else:
            return loader(file, n_polygons)



from openride import BoundingBox, Point, Rotation, Size, Transform
from openride.core.numba import transform_vertices
from openride.test.random_core_generator import get_random
from openride.test import get_pytest_benchmark

from shapely import geometry

import numpy as np
import pytest


pytest_benchmark = get_pytest_benchmark(group="core.bounding_box")


def test_init():
    b = BoundingBox()
    assert b.position == Point()
    assert b.rotation == Rotation()
    assert b.size == Size()


@pytest_benchmark
def test_box_get_bird_eye_view_vertices(benchmark):
    box = BoundingBox()
    vertices = benchmark(box.get_bird_eye_view_vertices)
    assert vertices.shape == (4, 2)


def test_bird_eye_view_vertices_translation():
    v = BoundingBox(Point(1, 2, 3)).get_bird_eye_view_vertices()
    assert np.all(v == np.array([[2, 3], [2, 1], [0, 1], [0, 3]]))


def test_bird_eye_view_vertices_rotation():
    v = BoundingBox(rotation=Rotation(0, 0, np.pi / 6)).get_bird_eye_view_vertices()
    x = np.sin(np.deg2rad(15)) * 2**0.5
    assert np.all(pytest.approx(v) == np.array([[x, 1 + x], [1 + x, -x], [-x, -1 - x], [-1 - x, x]]))


def test_bird_eye_view_vertices_scale():
    v = BoundingBox(size=Size(2, 2, 2)).get_bird_eye_view_vertices()
    assert np.all(v == np.array([[2, 2], [2, -2], [-2, -2], [-2, 2]]))


@pytest_benchmark
def test_box_get_vertices(benchmark):
    box = BoundingBox()
    vertices = benchmark(box.get_vertices)
    assert vertices.shape == (8, 3)


def test_bounding_box_to_shapely():
    b = BoundingBox()
    assert isinstance(b.to_shapely(), geometry.Polygon)


def test_get_transform():
    b = BoundingBox()
    assert isinstance(b.get_transform(), Transform)


def test_box_transform_identity():
    b = BoundingBox()
    assert b == b.transform(Transform())


def test_box_rotation():
    b = BoundingBox()
    tf = Transform(rotation=Rotation(0, 0, np.pi / 6))
    assert pytest.approx(b.transform(tf).rotation.yaw) == np.pi / 6


@pytest_benchmark
def test_box_transform(benchmark):
    box = get_random(BoundingBox)
    tf = Transform(Point(*np.random.random(3) * 10), Rotation(0, 0, np.random.random() * 2 * np.pi))
    benchmark(box.transform, tf)
    v1 = box.transform(tf).get_vertices()
    v2 = transform_vertices(box.get_vertices(), tf.get_matrix())
    assert np.all(pytest.approx(v1) == v2)



from openride import BoundingBox, Point, Polygon, Polyline, Rotation, Size, Transform

import numpy as np


def get_random(cls):

    if cls == Point:
        return Point(*np.random.random(3) * 10)

    elif cls == BoundingBox:
        return BoundingBox(
            Point(*np.random.random(3) * 10),
            Rotation(0, 0, np.random.random() * 2 * np.pi),
            Size(*np.random.random(3) * 10),
        )

    elif cls == Polyline:
        return Polyline(np.random.random((10, 3)))

    elif cls == Polygon:
        right_side = [Point(x + np.random.random() * 0.3, np.random.random() - 5) for x in range(10)]
        left_side = [Point(10 - x - np.random.random() * 0.3, np.random.random() + 5) for x in range(10)]
        return Polygon(right_side + left_side)

    elif cls == Transform:
        return Transform(
            Point(*np.random.random(3) * 10),
            Rotation(0, 0, np.random.random() * 2 * np.pi),
        )



from openride.core.geometry import Geometry
from openride.core.numba import transform_point

from dataclasses import dataclass
from shapely import geometry
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from openride.core.transform import Transform


@dataclass
class Point(Geometry):

    x: float = 0.0
    y: float = 0.0
    z: float = 0.0

    def to_shapely(self) -> geometry.Point:
        return geometry.Point([self.x, self.y, self.z])

    def transform(self, transform: "Transform") -> "Point":
        return Point(*transform_point(self.x, self.y, self.z, transform.get_matrix()))

    def __add__(self, other: "Point") -> "Point":
        return Point(self.x + other.x, self.y + other.y, self.z + other.z)

    def __sub__(self, other: "Point") -> "Point":
        return Point(self.x - other.x, self.y - other.y, self.z - other.z)

    def __mul__(self, factor: float) -> "Point":
        return Point(factor * self.x, factor * self.y, factor * self.z)

    def __rmul__(self, factor: float) -> "Point":
        return self * factor

    def __repr__(self):
        return f"Point(x:{self.x:.2f}, y:{self.y:.2f}, z:{self.z:.2f})"



from openride import PointCloud, Point, Rotation, Transform
from openride.test import get_pytest_benchmark

import numpy as np
import pytest


pytest_benchmark = get_pytest_benchmark(group="core.collections.point_cloud")


def test_init_empty():
    pc = PointCloud()
    assert len(pc) == 0


def test_init():
    pc = PointCloud([Point(), Point()])
    assert len(pc) == 2


def test_append():
    pc = PointCloud([Point(), Point()])
    pc.append(Point())
    assert len(pc) == 3


def test_extend():
    pc = PointCloud([Point(), Point()])
    pc2 = PointCloud([Point(), Point()])
    pc.extend(pc2)
    assert len(pc) == 4


def test_pop():
    pc = PointCloud([Point(), Point(1), Point()])
    point = pc.pop(1)
    assert point.x == 1
    assert len(pc) == 2


def test_repr():
    pc = PointCloud([Point(), Point()])
    assert isinstance(str(pc), str)


@pytest_benchmark
def test_getitem(benchmark):
    pc = PointCloud([Point(), Point(1)])
    point = benchmark(pc.__getitem__, 1)
    assert point.x == 1


@pytest_benchmark
def test_transform(benchmark):
    pc = PointCloud([Point() for _ in range(10000)])
    tf = Transform(Point(1), Rotation(0, 0, np.pi))
    pc = benchmark(pc.transform, tf)
    assert pytest.approx(pc[0].x) == -1


def test_get_point_cloud():
    pc = PointCloud([Point(), Point()])
    xyz = pc.get_point_cloud()
    assert xyz.shape == (2, 3)
    assert np.all(xyz == 0.0)



from openride.core.geometry import Geometry
from openride.core.point import Point
from openride.core.rotation import Rotation
from openride.core.size import Size
from openride.core.transform import Transform
from openride.core.numba import (
    bird_eye_view_box_vertices,
    box_vertices,
)

from dataclasses import dataclass
from shapely import geometry

import numpy as np


@dataclass
class BoundingBox(Geometry):

    position: Point = Point()
    rotation: Rotation = Rotation()
    size: Size = Size()

    def to_shapely(self) -> geometry.Polygon:
        return geometry.Polygon(self.get_bird_eye_view_vertices())

    def transform(self, transform: "Transform") -> "BoundingBox":
        return BoundingBox(
            self.position.transform(transform),
            self.rotation + transform.rotation,
            self.size,
        )

    def get_transform(self) -> Transform:
        return Transform(self.position, self.rotation)

    def get_bird_eye_view_vertices(self) -> np.ndarray:
        return bird_eye_view_box_vertices(
            self.position.x,
            self.position.y,
            self.size.x,
            self.size.y,
            self.rotation.yaw,
        )

    def get_vertices(self) -> np.ndarray:
        return box_vertices(
            self.position.x,
            self.position.y,
            self.position.z,
            self.size.x,
            self.size.y,
            self.size.z,
            self.rotation.roll,
            self.rotation.pitch,
            self.rotation.yaw,
        )



from openride.core.bounding_box import BoundingBox
from openride.core.geometry import Geometry
from openride.core.point import Point
from openride.core.polygon import Polygon
from openride.core.polyline import Polyline

from openride.core.numba import (
    bev_distance_between_box_and_point,
    bev_distance_between_boxes,
    bev_distance_between_line_and_box,
    bev_distance_between_line_and_point,
    bev_distance_between_line_and_polygon,
    bev_distance_between_lines,
    bev_distance_between_points,
    bev_distance_between_polygon_and_point,
    bev_distance_between_polygons,
    bev_distance_between_polygon_and_box,
)

from multipledispatch import dispatch


@dispatch(Point, Point)
def bird_eye_view_distance(obj1: Point, obj2: Point) -> float:
    return bev_distance_between_points(obj1.x, obj1.y, obj2.x, obj2.y)


@dispatch(Point, BoundingBox)
def bird_eye_view_distance(obj1: Point, obj2: BoundingBox) -> float:
    return bev_distance_between_box_and_point(
        obj2.position.x,
        obj2.position.y,
        obj2.size.x,
        obj2.size.y,
        obj2.rotation.yaw,
        obj1.x,
        obj1.y,
    )


@dispatch(Point, Polyline)
def bird_eye_view_distance(obj1: Point, obj2: Polyline) -> float:
    return bev_distance_between_line_and_point(obj2.vertices, obj1.x, obj1.y)


@dispatch(Point, Polygon)
def bird_eye_view_distance(obj1: Point, obj2: Polygon) -> float:
    return bev_distance_between_polygon_and_point(obj2.vertices, obj1.x, obj1.y)


@dispatch(BoundingBox, Point)
def bird_eye_view_distance(obj1: BoundingBox, obj2: Point) -> float:
    return bev_distance_between_box_and_point(
        obj1.position.x,
        obj1.position.y,
        obj1.size.x,
        obj1.size.y,
        obj1.rotation.yaw,
        obj2.x,
        obj2.y,
    )


@dispatch(BoundingBox, BoundingBox)
def bird_eye_view_distance(obj1: BoundingBox, obj2: BoundingBox) -> float:
    return bev_distance_between_boxes(
        obj1.position.x,
        obj1.position.y,
        obj1.size.x,
        obj1.size.y,
        obj1.rotation.yaw,
        obj2.position.x,
        obj2.position.y,
        obj2.size.x,
        obj2.size.y,
        obj2.rotation.yaw,
    )


@dispatch(BoundingBox, Polyline)
def bird_eye_view_distance(obj1: BoundingBox, obj2: Polyline) -> float:
    return bev_distance_between_line_and_box(
        obj2.vertices,
        obj1.position.x,
        obj1.position.y,
        obj1.size.x,
        obj1.size.y,
        obj1.rotation.yaw,
    )


@dispatch(BoundingBox, Polygon)
def bird_eye_view_distance(obj1: BoundingBox, obj2: Polygon) -> float:
    return bev_distance_between_polygon_and_box(
        obj2.vertices,
        obj1.position.x,
        obj1.position.y,
        obj1.size.x,
        obj1.size.y,
        obj1.rotation.yaw,
    )


@dispatch(Polyline, Point)
def bird_eye_view_distance(obj1: Polyline, obj2: Point) -> float:
    return bev_distance_between_line_and_point(obj1.vertices, obj2.x, obj2.y)


@dispatch(Polyline, BoundingBox)
def bird_eye_view_distance(obj1: Polyline, obj2: BoundingBox) -> float:
    return bev_distance_between_line_and_box(
        obj1.vertices,
        obj2.position.x,
        obj2.position.y,
        obj2.size.x,
        obj2.size.y,
        obj2.rotation.yaw,
    )


@dispatch(Polyline, Polyline)
def bird_eye_view_distance(obj1: Polyline, obj2: Polyline) -> float:
    return bev_distance_between_lines(obj1.vertices, obj2.vertices)


@dispatch(Polyline, Polygon)
def bird_eye_view_distance(obj1: Polyline, obj2: Polygon) -> float:
    return bev_distance_between_line_and_polygon(obj1.vertices, obj2.vertices)


@dispatch(Polygon, Point)
def bird_eye_view_distance(obj1: Polygon, obj2: Point) -> float:
    return bev_distance_between_polygon_and_point(obj1.vertices, obj2.x, obj2.y)


@dispatch(Polygon, BoundingBox)
def bird_eye_view_distance(obj1: Polygon, obj2: BoundingBox) -> float:
    return bev_distance_between_polygon_and_box(
        obj1.vertices,
        obj2.position.x,
        obj2.position.y,
        obj2.size.x,
        obj2.size.y,
        obj2.rotation.yaw,
    )


@dispatch(Polygon, Polyline)
def bird_eye_view_distance(obj1: Polygon, obj2: Polyline) -> float:
    return bev_distance_between_line_and_polygon(obj2.vertices, obj1.vertices)


@dispatch(Polygon, Polygon)
def bird_eye_view_distance(obj1: Polygon, obj2: Polygon) -> float:
    return bev_distance_between_polygons(obj1.vertices, obj2.vertices)


# Default, fall back to shapely module if we don't have an implementation above
@dispatch(Geometry, Geometry)
def bird_eye_view_distance(obj1: Geometry, obj2: Geometry) -> float:
    return obj1.to_shapely().distance(obj2.to_shapely())



from openride import BoundingBox, Point, Polyline, Polygon, distance
from openride.core.numba import vertices_inside_box

from multipledispatch import dispatch

import numpy as np


@dispatch(BoundingBox, Point)
def contains(obj1: BoundingBox, obj2: Point) -> bool:
    return len(vertices_inside_box(np.array([(obj2.x, obj2.y, obj2.z)], np.float64), obj1.get_vertices())) == 1


@dispatch(BoundingBox, BoundingBox)
def contains(obj1: BoundingBox, obj2: BoundingBox) -> bool:
    return len(vertices_inside_box(obj2.get_vertices(), obj1.get_vertices())) == 8


@dispatch(BoundingBox, Polyline)
def contains(obj1: BoundingBox, obj2: Polyline) -> bool:
    return len(vertices_inside_box(obj2.vertices, obj1.get_vertices())) == len(obj2)


@dispatch(BoundingBox, Polygon)
def contains(obj1: BoundingBox, obj2: Polygon) -> bool:
    return len(vertices_inside_box(obj2.vertices, obj1.get_vertices())) == obj2.vertices.shape[0]


@dispatch(Polyline, Point)
def contains(obj1: Polyline, obj2: Point) -> bool:
    return distance(obj1, obj2) == 0.0


@dispatch(Polyline, Polyline)
def contains(obj1: Polyline, obj2: Polyline) -> bool:
    return all([contains(obj1, point) for point in obj2])



from openride.core.point import Point
from openride.core.rotation import Rotation
from openride.core.numba import transform_matrix, transform_inverse_matrix

from dataclasses import dataclass

import numpy as np


@dataclass
class Transform:

    translation: Point = Point(0, 0, 0)
    rotation: Rotation = Rotation(0, 0, 0)

    def get_matrix(self) -> np.ndarray:
        return transform_matrix(
            self.translation.x,
            self.translation.y,
            self.translation.z,
            self.rotation.roll,
            self.rotation.pitch,
            self.rotation.yaw,
        )

    def get_inverse_matrix(self) -> np.ndarray:
        return transform_inverse_matrix(
            self.translation.x,
            self.translation.y,
            self.translation.z,
            self.rotation.roll,
            self.rotation.pitch,
            self.rotation.yaw,
        )

    def inverse(self) -> "Transform":
        return Transform.from_matrix(self.get_inverse_matrix())

    @classmethod
    def from_matrix(cls, matrix: np.ndarray) -> "Transform":
        return cls(Point(*matrix[:3, 3]), Rotation.from_matrix(matrix))



from abc import ABC, abstractclassmethod
from shapely.geometry.base import BaseGeometry
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from openride.core.transform import Transform


class Geometry(ABC):
    @abstractclassmethod
    def to_shapely(self) -> BaseGeometry:
        """Converts in a shapely.BaseGeometry object"""

    @abstractclassmethod
    def transform(self, transform: "Transform") -> "Geometry":
        """Returns a transformed version of this object"""



from openride import Category, SubCategory, Rotation

from typing import List, Union

import glob
import os
import numpy as np

FILEPATH = os.path.dirname(os.path.abspath(__file__))


DEFAULT_MODELS = {
    Category.Vehicle: f"{FILEPATH}/cars/Aston_martin.stl",
    Category.Pedestrian: f"{FILEPATH}/pedestrians/default_male.stl",
    Category.Cyclist: f"{FILEPATH}/cyclists/bicycle2.stl",
    Category.TrafficCone: f"{FILEPATH}/traffic_cones/cone.stl",
    SubCategory.Car: f"{FILEPATH}/cars/Aston_martin.stl",
    SubCategory.Van: f"{FILEPATH}/vans/volkswagen_transporter.stl",
    SubCategory.Bus: f"{FILEPATH}/buses/ShortSchoolBus.stl",
    SubCategory.Truck: f"{FILEPATH}/trucks/Truck2.stl",
    SubCategory.Motorcycle: f"{FILEPATH}/motorcycles/kawaski.stl",
}


DIRECTORIES = {
    Category.Vehicle: f"{FILEPATH}/cars",
    Category.Pedestrian: f"{FILEPATH}/pedestrians",
    Category.Cyclist: f"{FILEPATH}/cyclists",
    Category.TrafficCone: f"{FILEPATH}/traffic_cones",
    SubCategory.Car: f"{FILEPATH}/cars",
    SubCategory.Van: f"{FILEPATH}/vans",
    SubCategory.Bus: f"{FILEPATH}/buses",
    SubCategory.Truck: f"{FILEPATH}/trucks",
    SubCategory.Motorcycle: f"{FILEPATH}/motorcycles",
}


def get_model_files(category: Union[Category, SubCategory]) -> List[str]:
    directory = DIRECTORIES.get(category)
    if not directory:
        return []
    return glob.glob(f"{directory}/*")


MODEL_ROTATIONS = {
    "jeep_renegate.stl": Rotation(yaw=np.pi / 2),
    "Aston_martin.stl": Rotation(yaw=np.pi),
    "chevrolet_camaro_ss.stl": Rotation(yaw=np.pi / 2),
    "chevrolet_corvette_zr1.stl": Rotation(yaw=np.pi / 2),
    "dodge_challenger.stl": Rotation(yaw=np.pi / 2),
    "mitsubishi_lancer_evolution.stl": Rotation(yaw=np.pi / 2),
    "nissan_skyline_gtr.stl": Rotation(yaw=np.pi / 2),
    "shelby_mustang_gt350r.stl": Rotation(yaw=np.pi / 2),
    "tesla_model_3.stl": Rotation(yaw=np.pi / 2),
    "volkswagen_golf_gti.stl": Rotation(yaw=np.pi / 2),
    "chopper_reduziert.stl": Rotation(roll=np.pi / 2),
    "ducati.stl": Rotation(yaw=np.pi / 2),
    "kawaski.stl": Rotation(yaw=np.pi / 2),
    "volkswagen_transporter.stl": Rotation(yaw=np.pi / 2),
    "vw_transporter2.stl": Rotation(yaw=np.pi / 2),
    "Fire_Truck.stl": Rotation(),
    "trashtruck.stl": Rotation(),
    "Truck2.stl": Rotation(),
    "ShortSchoolBus.stl": Rotation(),
    "default_male.stl": Rotation(roll=np.pi / 4, pitch=0, yaw=np.pi / 2),
    "sophie_walk.stl": Rotation(roll=np.pi / 2, yaw=np.pi / 2),
    "tyler_walk.stl": Rotation(roll=np.pi / 2, yaw=np.pi / 2),
    "woman_stand.stl": Rotation(yaw=np.pi / 2),
    "default_female.stl": Rotation(yaw=np.pi / 2),
    "vahlen_stand.stl": Rotation(yaw=np.pi / 2),
    "bicycle.stl": Rotation(),
    "bicycle2.stl": Rotation(),
}



from enum import Enum


class Category(Enum):
    Unknown = 0
    Vehicle = 1
    Pedestrian = 2
    Cyclist = 3
    TrafficCone = 4
    Trailer = 5
    Other = 6


class SubCategory(Enum):
    Unknown = 0
    Car = 1
    Van = 2
    Bus = 3
    Truck = 4
    Motorcycle = 5


SUBCATEGORY_RELATIONS = {c: [SubCategory.Unknown] for c in Category}

SUBCATEGORY_RELATIONS[Category.Vehicle] += [
    SubCategory.Car,
    SubCategory.Van,
    SubCategory.Bus,
    SubCategory.Truck,
    SubCategory.Motorcycle,
]



from openride.core.geometry import Geometry
from openride.core.point import Point
from openride.core.numba import (
    linear_interpolation,
    distance_to_index,
    index_to_distance,
    resample_linear,
    tangent_angles,
    radius_of_curvature,
    vertices_to_distances,
    transform_vertices,
)

from dataclasses import dataclass
from shapely import geometry
from typing import Tuple, Union, List, TYPE_CHECKING

import copy
import numpy as np

if TYPE_CHECKING:
    from openride.core.transform import Transform


@dataclass
class Polyline(Geometry):

    vertices: Union[np.ndarray, List[Point]] = np.empty((0, 3))

    def __post_init__(self):

        if isinstance(self.vertices, list):
            points = copy.deepcopy(self.vertices)
            self.vertices = np.zeros((len(points), 3))
            for i, point in enumerate(points):
                self.vertices[i] = point.x, point.y, point.z

    def to_shapely(self) -> Union[geometry.LineString, geometry.Point]:
        if len(self) > 1:
            return geometry.LineString(self.vertices[:, :2])
        elif len(self) == 1:
            return geometry.Point(self.vertices[0, :2])

    def transform(self, transform: "Transform") -> "Polyline":
        return Polyline(transform_vertices(self.vertices, transform.get_matrix()))

    def __len__(self) -> int:
        return self.vertices.shape[0]

    def append(self, point: Point):
        self.vertices = np.vstack([self.vertices, np.array([point.x, point.y, point.z])])

    def extent(self, polyline: "Polyline"):
        self.vertices = np.vstack([self.vertices, polyline.vertices])

    def get_distances(self, bird_eye_view: bool = False) -> np.ndarray:
        return vertices_to_distances(self.vertices, bird_eye_view)

    def get_total_distance(self, bird_eye_view: bool = False) -> float:
        return self.get_distances(bird_eye_view)[-1]

    def index_to_distance(self, index: float) -> float:
        return index_to_distance(self.vertices, index)

    def distance_to_index(self, distance: float) -> float:
        return distance_to_index(self.vertices, distance)

    def __getitem__(self, key: Union[slice, float, int]) -> Union["Polyline", Point]:

        if isinstance(key, slice):

            if isinstance(key.start, float):
                start = self[key.start]
                return Polyline(
                    np.vstack(
                        [
                            np.array([[start.x, start.y, start.z]]),
                            self[int(key.start + 1) : key.stop : key.step].vertices,
                        ]
                    )
                )

            if isinstance(key.stop, float):
                stop = self[key.stop]
                return Polyline(
                    np.vstack(
                        [
                            self[key.start : int(key.stop + 1) : key.step].vertices,
                            np.array([[stop.x, stop.y, stop.z]]),
                        ]
                    )
                )

            return Polyline(self.vertices[key])

        if key % 1 == 0.0:
            key = int(key)

        if isinstance(key, (int, np.int64)):
            return Point(*self.vertices[key])

        elif isinstance(key, float):
            x = linear_interpolation(self.vertices[:, 0], key)
            y = linear_interpolation(self.vertices[:, 1], key)
            z = linear_interpolation(self.vertices[:, 2], key)
            return Point(x, y, z)

    def get_tangent_angles(self) -> np.ndarray:
        return tangent_angles(self.vertices)

    def get_radius_of_curvature(self) -> np.ndarray:
        return radius_of_curvature(self.vertices)

    def resample(self, delta: float, return_indices: bool = False) -> Union["Polyline", Tuple["Polyline", np.ndarray]]:
        N = int(self.get_total_distance() / delta + 0.5) + 1
        resampled_vertices, indices = resample_linear(self.vertices, N)
        resampled = Polyline(resampled_vertices)
        if return_indices:
            return resampled, indices
        return resampled



from openride import Rotation

import numpy as np
import pytest


def test_init():
    r = Rotation(1, 2, 3)
    assert r.roll == 1 and r.pitch == 2 and r.yaw == 3


def test_post_init():
    r = Rotation(1 + 2 * np.pi, 2 - 2 * np.pi, 3 + 10 * np.pi)
    assert r.roll == 1 and r.pitch == 2 and r.yaw == 3


def test_rotation_matrix_identity():
    r = Rotation(0, 0, 0)
    assert np.all(r.get_matrix() == np.eye(3))


def test_rotation_from_identity():
    r = Rotation.from_matrix(np.eye(3))
    assert r.roll == 0 and r.pitch == 0 and r.yaw == 0


def test_rotation_euler_matrix_reciprocal():
    for _ in range(10):
        r1 = Rotation(*np.random.random(3) * 2 * np.pi)
        r2 = Rotation.from_matrix(r1.get_matrix())
        assert np.all(pytest.approx(r1.get_matrix()) == r2.get_matrix())


def test_add_rotations():
    r = Rotation(1, 2, 3) + Rotation(0.1, 0.2, 0.3)
    assert r.roll == 1.1
    assert r.pitch == 2.2
    assert r.yaw == 3.3


def test_sub_rotations():
    r = Rotation(1, 2, 3) - Rotation(0.1, 0.2, 0.3)
    assert r.roll == 0.9
    assert r.pitch == 1.8
    assert r.yaw == 2.7



from setuptools import setup, find_packages

with open("README.md", "r") as file:
    long_description = file.read()

requirements = [
    "shapely",
    "numba",
    "numpy",
    "dataclasses",
    "vtk>=9.1.0",
    "matplotlib",
    "multipledispatch",
]

setup(
    name="openride",
    version="0.1.1",
    author="Jean-Luc Déziel",
    author_email="jluc1011@hotmail.com",
    url="https://gitlab.com/jldez/openride",
    description="",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    install_requires=requirements,
)



from openride import Point
from openride.core.numba import cartesian_to_spherical, spherical_to_cartesian

from dataclasses import dataclass
from enum import Enum
from typing import Callable, Tuple, Union

import numpy as np
import vtk


class Event(Enum):
    MouseMoveEvent = "MouseMoveEvent"
    LeftButtonPressEvent = "LeftButtonPressEvent"
    MiddleButtonPressEvent = "MiddleButtonPressEvent"
    RightButtonPressEvent = "RightButtonPressEvent"
    LeftButtonReleaseEvent = "LeftButtonReleaseEvent"
    MiddleButtonReleaseEvent = "MiddleButtonReleaseEvent"
    RightButtonReleaseEvent = "RightButtonReleaseEvent"
    MouseWheelForwardEvent = "MouseWheelForwardEvent"
    MouseWheelBackwardEvent = "MouseWheelBackwardEvent"
    EnterEvent = "EnterEvent"
    LeaveEvent = "LeaveEvent"
    ExitEvent = "ExitEvent"
    KeyPressEvent = "KeyPressEvent"
    KeyReleaseEvent = "KeyReleaseEvent"


@dataclass
class MouseEvent:
    position: Tuple[int, int]
    last_position: Tuple[int, int]


@dataclass
class KeyEvent:
    key: str


class Interactor:
    def __init__(self, vtk_viewer, mouse_camera_interactions: bool = True):

        self.viewer = vtk_viewer

        self._interactor = self.viewer.renderWindow.MakeRenderWindowInteractor()

        # Set an empty style, so this interactor does nothing when we call ProcessEvents() except our callbacks.
        self._interactor.SetInteractorStyle(vtk.vtkInteractorEventRecorder())
        self._interactor.Initialize()

        self._interactor.SetLightFollowCamera(True)

        # Callbacks
        self.set_callback(Event.MouseMoveEvent, self.__on_move)
        self.set_callback(Event.LeftButtonPressEvent, self.__on_left_click)
        self.set_callback(Event.MiddleButtonPressEvent, self.__on_middle_click)
        self.set_callback(Event.RightButtonPressEvent, self.__on_right_click)
        self.set_callback(Event.LeftButtonReleaseEvent, self.__on_left_release)
        self.set_callback(Event.MiddleButtonReleaseEvent, self.__on_middle_release)
        self.set_callback(Event.RightButtonReleaseEvent, self.__on_right_release)
        self.set_callback(Event.MouseWheelForwardEvent, self.__on_scroll_up)
        self.set_callback(Event.MouseWheelBackwardEvent, self.__on_scroll_down)
        self.set_callback(Event.EnterEvent, self.__on_enter)
        self.set_callback(Event.LeaveEvent, self.__on_leave)
        self.set_callback(Event.ExitEvent, self.__on_exit)

        self.set_mouse_camera_interactions(mouse_camera_interactions)

        self._point_placer = vtk.vtkFocalPlanePointPlacer()

        self.cursor = (0, 0)
        self.cursor_is_in_window = True
        self.left_mouse = False
        self.middle_mouse = False
        self.right_mouse = False

    def toggle(self, state: bool):
        if state:
            self._interactor.Enable()
        else:
            self._interactor.Disable()

    def set_mouse_camera_interactions(self, state: bool):
        self._mouse_camera_interactions = state

    def update(self):
        self._interactor.ProcessEvents()

    def set_callback(self, event: Event, callback: Callable[[Union[MouseEvent, KeyEvent]], None]):

        if event is Event.KeyPressEvent or event is Event.KeyReleaseEvent:

            def _callback(key, _):
                callback(KeyEvent(key.GetKeySym()))

        else:

            def _callback(mouse, _):
                callback(MouseEvent(mouse.GetEventPosition(), mouse.GetLastEventPosition()))

        self._interactor.AddObserver(event.value, _callback, 1.0)

    def __on_enter(self, *args):
        self.cursor_is_in_window = True

    def __on_leave(self, *args):
        self.cursor_is_in_window = False

    def __on_left_click(self, *args):
        self.left_mouse = True
        self.grab_position = self.get_cursor_in_world()

    def __on_left_release(self, *args):
        self.left_mouse = False

    def __on_middle_click(self, *args):
        self.middle_mouse = True

    def __on_middle_release(self, *args):
        self.middle_mouse = False

    def __on_right_click(self, *args):
        self.right_mouse = True

    def __on_right_release(self, *args):
        self.right_mouse = False

    def __on_scroll_up(self, *args):
        self.zoom(1)

    def __on_scroll_down(self, *args):
        self.zoom(-1)

    def __on_exit(self, *args):
        exit()

    def __on_move(self, mouse_event: MouseEvent):
        if not self.cursor_is_in_window:
            return
        if not self._mouse_camera_interactions:
            return
        self.cursor = mouse_event.position

        position = self.viewer.camera.get_position()
        focus = self.viewer.camera.get_focus()

        if self.left_mouse:
            delta = self.get_cursor_in_world() - self.grab_position
            position -= delta
            focus -= delta

        if self.middle_mouse:
            x0, y0 = mouse_event.last_position
            (
                d,
                horizontal_angle,
                vertical_angle,
            ) = self.__get_camera_spherical_coordinates()
            factor = d / 1400
            delta_y = self.cursor[1] - y0
            position.z -= factor * delta_y
            focus.z -= factor * delta_y

        if self.right_mouse:
            x0, y0 = mouse_event.last_position
            delta_x = self.cursor[0] - x0
            delta_y = self.cursor[1] - y0

            (
                d,
                horizontal_angle,
                vertical_angle,
            ) = self.__get_camera_spherical_coordinates()
            horizontal_angle -= delta_x / 200
            vertical_angle = np.clip(vertical_angle + delta_y / 200, 0.001, np.pi - 0.001)
            delta = Point(*spherical_to_cartesian(d, horizontal_angle, vertical_angle))
            position = delta + focus

        self.viewer.camera.set_position(position)
        self.viewer.camera.set_focus(focus)

    def zoom(self, value: float):
        d, horizontal_angle, vertical_angle = self.__get_camera_spherical_coordinates()
        factor = d / 10
        d += value * factor
        delta = Point(*spherical_to_cartesian(d, horizontal_angle, vertical_angle))
        self.viewer.camera.set_position(delta + self.viewer.camera.get_focus())

    def __get_camera_spherical_coordinates(self) -> Tuple[float, float, float]:
        c = self.viewer.camera.get_position() - self.viewer.camera.get_focus()
        return cartesian_to_spherical(c.x, c.y, c.z)

    def get_cursor_in_world(self) -> Point:
        """Returns the intersection between the mouse cursor and the plane z = camera.focus"""
        x, y = self.cursor
        focal_plane_projection = self.__get_cursor_on_focal_plane_projection(x, y)
        return self.__focal_plane_projection_to_z_slice_projection(
            focal_plane_projection, self.viewer.camera.get_focus().z
        )

    def __get_cursor_on_focal_plane_projection(self, pixel_x: int, pixel_y: int) -> Point:
        world_position = [0, 0, 0]
        world_orientation = [0, 0, 0, 0, 0, 0, 0, 0, 0]
        self._point_placer.ComputeWorldPosition(
            self.viewer.renderer, (pixel_x, pixel_y), world_position, world_orientation
        )
        return Point(*world_position)

    def __focal_plane_projection_to_z_slice_projection(self, focal_plane_projection: Point, world_z: float) -> Point:
        cam_position = self.viewer.camera.get_position()
        cam_to_focal_proj = focal_plane_projection - cam_position
        r, theta, phi = cartesian_to_spherical(cam_to_focal_proj.x, cam_to_focal_proj.y, cam_to_focal_proj.z)
        r *= (world_z - cam_position.z) / cam_to_focal_proj.z
        cam_to_focal_proj = Point(*spherical_to_cartesian(r, theta, phi))
        return cam_position + cam_to_focal_proj



from openride.core.collections.collection import Collection
from openride.core.bounding_box import BoundingBox
from openride.core.point import Point
from openride.core.rotation import Rotation
from openride.core.size import Size
from openride.core.transform import Transform
from openride.core.numba import transform_vertices

from typing import List

import numpy as np


class BoundingBoxCollection(Collection):
    def __init__(self, bounding_boxes: List[BoundingBox] = []):

        self.dtype = np.dtype([("position", "f8", (3)), ("rotation", "f8", (3)), ("size", "f8", (3))])
        self._raw: np.ndarray = np.empty(0, self.dtype)

        for bounding_box in bounding_boxes:
            self.append(bounding_box)

    def __len__(self) -> int:
        return self._raw.shape[0]

    def append(self, bbox: BoundingBox):
        array = np.array(
            (
                (bbox.position.x, bbox.position.y, bbox.position.z),
                (bbox.rotation.roll, bbox.rotation.pitch, bbox.rotation.yaw),
                (bbox.size.x, bbox.size.y, bbox.size.z),
            ),
            self.dtype,
        )
        self._raw = np.hstack((self._raw, array))

    def extend(self, box_collection: "BoundingBoxCollection"):
        self._raw = np.hstack((self._raw, box_collection._raw))

    def pop(self, index: int) -> BoundingBox:
        bbox = self[index]
        self._raw = np.delete(self._raw, index, axis=0)
        return bbox

    def __getitem__(self, index: int) -> BoundingBox:
        return BoundingBox(
            position=Point(*self._raw[index]["position"]),
            rotation=Rotation(*self._raw[index]["rotation"]),
            size=Size(*self._raw[index]["size"]),
        )

    def transform(self, transform: Transform) -> "BoundingBoxCollection":
        tmp = np.copy(self._raw)
        tmp["position"] = transform_vertices(self._raw["position"], transform.get_matrix())
        tmp["rotation"][:, 0] = (tmp["rotation"][:, 0] + transform.rotation.roll) % (2 * np.pi)
        tmp["rotation"][:, 1] = (tmp["rotation"][:, 1] + transform.rotation.pitch) % (2 * np.pi)
        tmp["rotation"][:, 2] = (tmp["rotation"][:, 2] + transform.rotation.yaw) % (2 * np.pi)
        transformed = BoundingBoxCollection()
        transformed._raw = tmp
        return transformed

    def get_distances(self) -> np.ndarray:
        return (
            self._raw["position"][:, 0] ** 2 + self._raw["position"][:, 1] ** 2 + self._raw["position"][:, 2] ** 2
        ) ** 0.5

    def filter(self, indices: List[int]) -> "BoundingBoxCollection":
        filtered = BoundingBoxCollection()
        filtered._raw = self._raw[indices]
        return filtered

    def filter_distance(self, min_distance: float = 0.0, max_distance: float = 100.0) -> "BoundingBoxCollection":
        distances = self.get_distances()
        return self.filter(np.where((distances >= min_distance) & (distances <= max_distance)))



from openride import BoundingBox, BoundingBoxCollection, Point, Polyline, Polygon, PointCloud, Category, SubCategory
from openride.viewer import vtk_utils
from openride.viewer.camera import Camera
from openride.viewer.interactor import Interactor
from openride.viewer.hud import Hud
from openride.viewer.model_loader import ModelLoader
from openride.viewer.models import DEFAULT_MODELS

from typing import Iterable, List, Tuple, Union

import vtk


class Viewer:
    def __init__(
        self,
        resolution=(1600, 900),
        background: Tuple[float, float, float] = (0.05, 0.05, 0.1),
        mouse_camera_interactions: bool = True,
    ):

        self.renderer = vtk.vtkRenderer()
        self.renderWindow = vtk.vtkRenderWindow()
        self.renderWindow.AddRenderer(self.renderer)

        self.set_resolution(resolution)
        self.set_background(background)

        self.camera = Camera(self)
        self.interactor = Interactor(self, mouse_camera_interactions)
        self.hud = Hud(self, position=(50, 50))

        self._model_loader = ModelLoader()

        self._actors_to_clear = []
        self._text_actor_index = 0

    def set_resolution(self, resolution: Tuple[int, int]):
        self.renderWindow.SetSize(resolution[0], resolution[1])

    def set_background(self, color=Tuple[float, float, float]):
        self.renderer.SetBackground(*color)

    def update(self):
        self.interactor.update()
        self.hud.update()
        self.renderWindow.Render()
        self.__clear_actors()

    def render_actor(self, actor: vtk.vtkActor, persist: bool = False):
        self.renderer.AddActor(actor)
        if not persist:
            self._actors_to_clear.append(actor)

    def draw_point(
        self,
        point: Point,
        color: tuple = (1, 1, 1),
        size: int = 0.15,
        alpha: float = 1,
        resolution: int = 10,
        persist: bool = False,
    ):
        actor = vtk_utils.get_point_actor(point.x, point.y, point.z, color, size, alpha, resolution)
        self.render_actor(actor, persist)

    def draw_point_cloud(
        self,
        pcloud: PointCloud,
        color: tuple = None,
        colors: Iterable[float] = None,
        colormap: str = "jet",
        color_range: float = 2**16,
        size: int = 10,
        alpha: float = 1,
        persist: bool = False,
    ):
        if color is None and colors is None:
            color = (1, 1, 1)
        actor = vtk_utils.get_point_cloud_actor(
            pcloud.get_point_cloud(), colors, color, colormap, size, alpha, color_range
        )
        self.render_actor(actor, persist)

    def draw_bounding_box(
        self,
        bounding_box: Union[BoundingBox, BoundingBoxCollection],
        color: tuple = (1, 1, 1),
        alpha: float = 1,
        edge_color: tuple = None,
        persist: bool = False,
    ):

        if isinstance(bounding_box, BoundingBoxCollection):
            [self.draw_bounding_box(box, color, alpha, edge_color, persist) for box in bounding_box]
            return

        actor = vtk_utils.get_box_actor(
            bounding_box.position.x,
            bounding_box.position.y,
            bounding_box.position.z,
            bounding_box.size.x,
            bounding_box.size.y,
            bounding_box.size.z,
            bounding_box.rotation.roll,
            bounding_box.rotation.pitch,
            bounding_box.rotation.yaw,
            color,
            alpha,
        )

        if edge_color is not None:
            actor.GetProperty().SetEdgeVisibility(1)
            actor.GetProperty().SetLineWidth(10)
            actor.GetProperty().SetEdgeColor(*edge_color)

        self.render_actor(actor, persist)

    def draw_model(
        self,
        model: Union[str, Category, SubCategory],
        bounding_box: BoundingBox,
        number_polygons: int = 100000,
        color: tuple = (1, 1, 1),
        alpha: float = 1,
        persist: bool = False,
    ):
        p = bounding_box.position
        s = bounding_box.size
        r = bounding_box.rotation
        if isinstance(model, (Category, SubCategory)):
            file = DEFAULT_MODELS.get(model)
        else:
            file = model
        if not file:
            return
        actor = vtk_utils.get_model_actor(
            self._model_loader,
            file,
            number_polygons,
            p.x,
            p.y,
            p.z,
            s.x,
            s.y,
            s.z,
            r.roll,
            r.pitch,
            r.yaw,
            color,
            alpha,
        )
        self.render_actor(actor, persist)

    def draw_polyline(
        self,
        line: Polyline,
        color: tuple = (1, 1, 1),
        alpha: float = 1,
        linewidth: int = 2,
        persist: bool = False,
    ):
        actor = vtk_utils.get_line_actor(line.vertices, color, alpha, linewidth)
        self.render_actor(actor, persist)

    def draw_polylines(
        self,
        lines: List[Polyline],
        color: tuple = (1, 1, 1),
        alpha: float = 1,
        linewidth: int = 2,
        persist: bool = False,
    ):
        actor = vtk_utils.get_lines_actor(
            [line.vertices for line in lines],
            color=color,
            alpha=alpha,
            linewidth=linewidth,
        )
        self.render_actor(actor, persist)

    def draw_polygon(
        self,
        polygon: Polygon,
        color: tuple = (1, 1, 1),
        alpha: float = 1,
        persist: bool = False,
    ):
        actor = vtk_utils.get_polygon_actor(polygon.vertices, color=color, alpha=alpha)
        self.render_actor(actor, persist)

    def labelize(self, label: str, pos: Point, fontsize: int = 16, persist: bool = False):
        if not self.camera.is_in_front(pos):
            return
        actor = vtk_utils.get_text_actor(label, pos.x, pos.y, pos.z, fontsize, self._text_actor_index)
        self.render_actor(actor, persist)
        self._text_actor_index += 1

    def draw_arrow(
        self,
        start: Point,
        end: Point,
        color: tuple = (1, 1, 1),
        alpha: float = 1,
        linewidth: int = 10,
        headsize: float = 1,
        persist: bool = False,
    ):
        self.draw_polyline(Polyline([start, end]), color=color, alpha=alpha, linewidth=linewidth)
        actor = vtk_utils.get_cone_actor(
            start.x,
            start.y,
            start.z,
            end.x,
            end.y,
            end.z,
            color,
            alpha,
            headsize,
        )
        self.render_actor(actor, persist)

    def draw_grid(
        self,
        origin: Point = Point(),
        delta: float = 5,
        extent: float = 200,
        alpha: float = 1.0,
        persist: bool = True,
    ):
        lines = []

        for s in [0, 1]:
            for i in range(int(2 * extent / delta) + 1):
                points = []
                if s == 0:
                    points.append(Point(-extent, -extent + i * delta) + origin)
                    points.append(Point(extent, -extent + i * delta) + origin)
                else:
                    points.append(Point(-extent + i * delta, -extent) + origin)
                    points.append(Point(-extent + i * delta, extent) + origin)
                lines.append(Polyline(points))

        self.draw_polylines(lines, color=(0.3, 0.3, 0.3), alpha=alpha, linewidth=1, persist=persist)

    def __clear_actors(self):
        for actor in self._actors_to_clear:
            self.renderer.RemoveActor(actor)
        self._actors_to_clear = []
        self._text_actor_index = 0

    def close(self):
        del self.renderWindow
        self.interactor._interactor.InvokeEvent("ExitEvent")
